/*
 * LCD_Config v2.1
 * ---------------
 *
 * last change: 12.02.2011
 *
 * Allows configuration of the MultiWii software via a serial connection.
 * Requires a LCD display and at least four buttons (left, right, up, down)
 * or an analog joystick.
 *
 * Options:
 *   Logging functionality (requires a SD card), output to binary or LogView OpenFormat.
 *   Status LEDs (three LEDs, two are sufficient without logging)
 *   Setting of LCD Backlight intensity
 *   Use of ShiftRegLCD lib instead of LiquidChrystal lib
 *   Hardware buttons instead of buttons connected via cascading resistors
 *   Analog joystick instead of buttons
 *   Graphic output (simple character graphic for angles and servo signals)
 *   MinMax mode (store und display minimum/maximum values of certain fields) 
 *   [Test] (for standalone testing without a MWC board connected)
 *
 * IMPORTANT: undefining the LOGGING function does not work correctly. When commenting
 * out the line "#include <SD.h>", the image size decreases by about 4k. Recommended
 * in case you are targeting a ATMega168.
 *
 * ----------
 * Graphic Mode is toggled either by pressing the SELECT key or by pressing the RIGHT key longer
 * than 0.5 sec. Min/Max Mode is toggled by pressing the RIGHT key shorter than 0.5 sec.
 * ----------
 * For infos on LogView fileformat, see www.logview.info
 * To import logfiles into LogView, you need to copy the attached .ini and .jpg File
 * to the directory <username>/<application data>/LogView/OpenFormat, select the
 * device LCD_CONFIG, and click on "import device file".
 * Data is formatted to three LogView channels: the first channel contains sensor data,
 * the second servo and motor PWM signals, the third the receiver channels. 
 * ----------
 *
 * Authors: yacco, calli
 *
 * changes to 1.25: 
 *   display values of gyro and acc changed to signed
 *   show minimum/maximum values of sensor data and servo signals
 * changes to 1.28:
 *   wait state after startup, no serial communication during initialization of the MWC
 *   removed min/max values
 *   added LED status signals
 * changes to 1.30:
 *   bugfix in playback routine, added timeout in playback
 * changes to 1.31:
 *   minor bugfixes, added setting of backlight
 * changes to 1.32:
 *   removed KEY_SELECT column in next_state, backlight and waitForInit optional
 * changes to 1.33:
 *   check trailing 'A' char in serial communication to detect frame errors, 
 *   optimization of format routines (merged two routines, invalid values are displayed as ###)
 *   bugfix in logging (logMode=-1 was stored in flash, so logging was permanently disabled)
 *   KEY_LEFT cancels reading when no serial connection is present (important to reset wrong settings!) 
 * changes to 1.5:
 *   reimplemented Min/Max-Mode (found it useful again), 
 *   fixed compiling problems for certain combinations of options 
 * changes to 1.6:
 *   adaptation to MWC version 1.preter6 (new config "Flying Wing", new data format)
 *   changed status display (state 0 - Config)
 *   bug 'logMode=-1' finally solved
 *   Version has been set to 2.0 to disambiguate between MWC version and LCD_Conf version.
 * changes to 2.0 beta: 
 *   swapped motors 4 and 5 in 'Esc Out' for versions 1.5 and 1.6prebis
 *   changed behaviour of UP and DOWN for consistency (UP decrements, DOWN increments values)
 *   added version V1.6 
 *   output logfiles in LogView format [OPTIONAL] (file numbers starting from 500, extension ".lvw")
 *   added JOYSTICK option to replace input buttons by an analog joystick [experimental]
 *   display 'invalid' values in graphics
 *   heavy optimizations (!) due to space restrictions
 *   removed WAIT_FOR_INIT option, changed start state to 31 instead 
 * changes to 2.0:
 *   bugfix in deserialize() with v1.6
 * changes to 2.01:
 *   Joystick handling was improved (calibration, use of pullups, can be used in parallel to buttons)
 *   Added 'Play LogFile' state
 *   Easter egg function to display internal configuration :)
 *   BUGFIX: RC Rate multiplied by factor 2 for MWC Versions >v1.5, minimum of 20% for RC Rate (0% makes no sense)
 *   Added random offsets in TEST mode
 * 
 * overall footprint (on an Arduino Uno, using the LiquidCrystal lib and cascading buttons): 
 *  ~28464 bytes with all options enabled (very close to the maximum image size uploadable via bootloader [~28650 bytes])
 *  ~12570 bytes with all options disabled (and //#include <SD.h>)
 * 
 * TO DO: 
 *  show pan-tilt servos, ...
 *
 */

// ----- begin of configuration parameters -----

// use ShiftRegLCD lib instead of LiquidCrystal lib
// [footprint: decreases ~200 bytes]
//#define SRLCD     

// ----------

// use Hardwarekeys connected to Ports
// [footprint: ~120 bytes]
//#define HWKEYS

// Input pins of the hardbuttons: (used when HWKEYS is defined)
#define HW_RIGHT    10
#define HW_UP        9
#define HW_DOWN     11
#define HW_LEFT      8
#define HW_SELECT    6

// ----------

// Use an analog Joystick (e.g., Sparkfun Thumb Joystick) instead of Keys (experimental!)
// Joystick and Buttons can be used simultaneously. The center position of the Joystick
// is calibrated during startup, so leave the stick in central position when switching on.
// [footprint: ~150 bytes]
//#define JOYSTICK

// Caution: Joystick Pins are in conflict with default LED Pins!
#define JS_SELECT_PIN A1
#define JS_HOR_PIN    A2
#define JS_VERT_PIN   A3

// Configuration of the Axes
#define JS_HOR_SIGN  +1 
#define JS_VERT_SIGN +1
#define JS_DEADBAND  32

// inverts the switch level of the SELECT key (default: high)
//#define JS_INV_SELECT_KEY

// ----------

// include logging to SD/SDHC card (formatted to FAT16 or FAT32) via SPI
// [footprint: ~12 kbytes]
#define LOGGING
#define CHIP_SELECT 10

// enables output in LogView format (~950 bytes)
#define LOGVIEW

// Logging Modes resp. LogFile Formats:
#define LOG_ERROR  -1
#define NO_LOGGING  0
#define LF_BINARY   1
#define LF_LOGVIEW  2

// ----------

// include routines to show some output as simple character graphic (for servo signals and angles) 
// [footprint: ~2.4 kbytes]
#define GRAPHIC

// ----------

// track minimum and maximum values of certain values
// [footprint: ~600 bytes]
#define MINMAX

#ifdef MINMAX
  #define MM_ACT [0]
  #define MM_VAL [minMaxMode]
#else
  #define MM_ACT
  #define MM_VAL
#endif

// ----------

// include routines to show the state of the application with three leds 
// (red: error, yellow: busy/waiting for input, green: logging)
// [footprint: ~250 bytes]
#define STATUS_LEDS

// Caution: Default Pins of LEDs and Joystick input overlap
#define RED_LED      A1
#define YELLOW_LED   A3
#define GREEN_LED    A5

// ----------

// if defined, allows dimming of LCD backlight intensity (Pin 10)
// [footprint: ~580 bytes]
#define BACKLIGHT

#define BACKLIGHT_PIN 10

// -----

// simulates a serial connection, SELECT key in state "ESC Out" changes the configuration
//#define TEST

// ----------------

// Number of hardbuttons:
#define NUM_KEYS     5

// Identifiers for the five hardbuttons:
#define KEY_RIGHT    0
#define KEY_UP       1
#define KEY_DOWN     2
#define KEY_LEFT     3
#define KEY_SELECT   4
#define STATUS       5 // Last column in the transition table is reserved for status flags

// RC Channel Aliases:
#define ROLL         0
#define PITCH        1
#define YAW          2
#define THROTTLE     3
#define AUX1         4

// ----- end of configuration parameters -----

// States:
#ifdef LOGGING
  #define NUM_STATES  47
#else
  #define NUM_STATES  42
#endif

// Number of Configurations: (Unknown, Tri, Quad+, QuadX, Bi, Gimbal, Y6, Hex, Flying Wing)
#define NUM_CONFIGS  9

// Speed of the serial interface. 115200 kbit/s is the best option for a USB connection.
#define SERIAL_COM_SPEED 115200

// ----------------

// for persistence of SW Version, Logging Mode, LogDelay and Profiles
#include <EEPROM.h>

// defines the PROGMEM macro to move constant variables from SRAM to flash memory
#include <avr/pgmspace.h> 

// Workaround for "insertion of forward declarations in excluded blocks" problem
// Don't move this line around.
void setup();     

#ifndef SRLCD
  #include <LiquidCrystal.h>
  LiquidCrystal lcd = LiquidCrystal(8, 9, 4, 5, 6, 7); // rs, enable, d4, d5, d6, d7
#else
  #include <ShiftRegLCD.h>
  ShiftRegLCD lcd(3, 2, 4, 2);        // (Datapin, Clockpin, Enablepin or TWO_WIRE [, Lines [, Font]]])
#endif

#ifdef LOGGING
  #include <SD.h>                     // ***** comment this line when LOGGING is not required - saves 4k of flash *****
  static File logFile;
  static char fileName[] = "000.log"; // LogFiles consists of a three digit number (000..999) with extension ".log"
#endif

// ----------------

PROGMEM prog_char metaData[] = "v2.1:" 
#ifdef TEST
  "T"
#endif
#ifdef SRLCD
  "S"
#endif
#ifdef HWKEYS
  "H"
#endif
#ifdef JOYSTICK
  "J"
#endif
#ifdef GRAPHIC
  "G"
#endif
#ifdef MINMAX
  "M"
#endif
#ifdef LOGGING
  "L"
#endif
#ifdef BACKLIGHT
  "B"
#endif
#ifdef STATUS_LEDS
  "$"
#endif
;

// ----------------

PROGMEM prog_char intro       [] = "LCD Config 2.1 ";
PROGMEM prog_char emptyLine   [] = "                ";
PROGMEM prog_char reading     [] = " Reading...     ";
PROGMEM prog_char writing     [] = " Writing...     ";
PROGMEM prog_char versMsg     [] = "MWC Version: ";
PROGMEM prog_char readMsg     [] = " Read 00 Bytes  ";
PROGMEM prog_char writeMsg    [] = " Wrote 00 Bytes ";
PROGMEM prog_char cancelRead  [] = "Reading canceled";
PROGMEM prog_char frameError  [] = "Inv. FrameLength";
PROGMEM prog_char timeoutError[] = " Timeout Error  ";

PROGMEM prog_char modeMsg  [][9] = { "AcroMode", "Lev.Mode", " Offline" };
PROGMEM prog_char accMsg      [] = "Acc";
PROGMEM prog_char statusMsg   [] = "Nck  Baro  Comp ";
PROGMEM prog_char yesNoMsg    [] = " \177: no, \176: yes  ";

PROGMEM prog_char haltMsg [][17] = { " Program halted ", " Press 'RESET'  " };
PROGMEM prog_char resetMsg    [] = "! Reset Values !";

#ifdef BACKLIGHT
PROGMEM prog_char bLightMsg   [] = "LCD Backlight:  ";
#endif

#ifdef LOGGING
PROGMEM prog_char logMsg      [] = " Logging...     ";
PROGMEM prog_char noCard      [] = "No SD Card found";
PROGMEM prog_char fileError   [] = "Open File failed";
PROGMEM prog_char fileDeleted [] = "Logfile deleted ";
PROGMEM prog_char deleteFailed[] = "Unable to delete";
PROGMEM prog_char haltReqMsg  [] = " Halt Program?  ";
PROGMEM prog_char logEnbMsg   [] = " Logging active ";
PROGMEM prog_char stopLogMsg  [] = " Stop Logging?  ";
//PROGMEM prog_char playMsg1    [] = " Play Logfile?  ";
//PROGMEM prog_char playMsg3    [] = "Select Logfile: ";
//PROGMEM prog_char playMsg4    [] = "Play Logfile:   ";
PROGMEM prog_char noFileMsg   [] = "No Logfile found";
PROGMEM prog_char fileMsg     [] = "File: ";
PROGMEM prog_char fileSuff    [] = "000 #00000";
  #ifdef LOGVIEW
PROGMEM prog_char formatMsg   [] = "Select Format:  ";
PROGMEM prog_char logFormats  [] = "\177:Bin, \176:LogView";
  #endif
#endif

#ifdef MINMAX
PROGMEM prog_char minMaxMsg   [] = "Mode:";
PROGMEM prog_char minMaxDsc[][4] = { "Act", "Min", "Max" };
#endif

//#ifdef GRAPHIC
//PROGMEM prog_char graphicMsg  [] = "Graphic:";
//PROGMEM prog_char onOffMsg [][4] = { "OFF", "ON " };
//#endif

// ----------------

#define V1_5        0
#define V1_6_prebis 1
#define V1_6        2

PROGMEM prog_char versions[][7] = { 
  "1.5   ", 
  "1.6pre", 
  "1.6   ",
};

// number of bytes per frame send from the MultiWii board, dependant on version
static uint8_t frameLength[] = { 
  65,  // 1.5
  70,  // 1.6prebis
  74,  // 1.6, 1.6preter
};

// configuration ids
PROGMEM prog_char configs[NUM_CONFIGS][8] = { 
  "Unknown", 
  "    Tri", 
  "  Quad+", 
  "  QuadX", 
  "     Bi", 
  " Gimbal", 
  "     Y6", 
  "    Hex", 
  "Fl.Wing",
};

// ----------------

//State messages
PROGMEM prog_char msgs[NUM_STATES][17] = {
//   0123456789012345 
    "1  Config       ", // 0
    "2  P-I-D        ", // 1
    "3  Rates        ", // 2
    "4  Sensors      ", // 3
    "5  Receiver In  ", // 4
    "6  ESC/Servo Out", // 5
    
    "Roll  P   I   D ", // 6
    "Pitch P   I   D ", // 7
    "Yaw   P   I   D ", // 8

    "D/R   Expo   ALS", // 9
    "R/P   Yaw   Thro", // 10
    
    "Gyr: Rol Pit Yaw", // 11
    "Acc: Rol Pit Z  ", // 12
    "Rho   Phi  Cycle", // 13
    
    "Rec: Rol Pit Yaw", // 14
    "Rec:  Thro   Aux", // 15

    "Lf: 000  Ri: 000", // 16 Tri
    "Ru: 000  Re: 000", // 17
    
    "Lf: 000  Ri: 000", // 18 Quad+
    "Fr: 000  Re: 000", // 19

    "FL: 000  FR: 000", // 20 QuadX
    "RL: 000  RR: 000", // 21

    "M1: 000  M2: 000", // 22 Bi
    "S1: 000  S2: 000", // 23

    " Roll:  000     ", // 24 Gimbal
    "Pitch:  000     ", // 25

    "Up 000  000  000", // 26 Y6 (upper - left rear right)
    "Dn 000  000  000", // 27    (lower - left rear right)

    "Fr 000  000  000", // 28 Hex (front - left mid right)
    "Rr 000  000  000", // 29     (rear  - left mid right)
    
    "Alt   Head      ", // 30
    
    "1  Config       ", // 31 - Acc, Bar, Mag
    
    "7  Profiles     ", // 32
    "8  Settings     ", // 33

    " Load Profile   ", // 34
    " Save Profile   ", // 35
    "Clear Profile   ", // 36

    " Load Profile:  ", // 37
    " Save Profile:  ", // 38
    "Clear Profile:  ", // 39

    "S1: 000  S2: 000", // 40 Flying Wing (servo 0, 1)
    "S3: 000  S4: 000", // 41             (servo 2, 3)

#ifdef LOGGING
    "9  Logging      ", // 42
    "Start Logging   ", // 43
    "Play Logfile    ", // 44
    "Delete Logfile  ", // 45
    "Delay Time      ", // 46
#endif
};

#ifdef LOGGING
  #define UP0 42
  #define DN0 42
#else
  #define UP0 33
  #define DN0 0
#endif

// state transition table, values above 80 are pseudo-states
PROGMEM prog_char nextState[NUM_STATES][NUM_KEYS] = {
  // Right Up  Down Left  Status      Status Flag: 1=Refresh Data, 2=2-Line Mode 
  {   85, UP0,   1,  31,   1, }, //  0 - Config
  {    6,   0,   2,   0,   0, }, //  1 - PID
  {    9,   1,   3,   0,   0, }, //  2 - Rates
  {   11,   2,   4,   0,   0, }, //  3 - Sensors
  {   14,   3,   5,   0,   0, }, //  4 - Rec In
  {   99,   4,  32,   0,   0, }, //  5 - ESc Out

  {   80,   8,   7,   1,   0, }, //  6 - PID Roll
  {   81,   6,   8,   1,   0, }, //  7 - PID Pitch
  {   82,   7,   6,   1,   0, }, //  8 - PID Yaw

  {   83,  10,  10,   2,   0, }, //  9 - DR, Expo, AutoLevel Strength
  {   84,   9,   9,   2,   0, }, // 10 - R/P Rate, Yaw Rate, Throttle PID

  {   93,  30,  12,   3,   1, }, // 11 - Gyro
  {   93,  11,  13,   3,   1, }, // 12 - Acc
  {   93,  12,  30,   3,   1, }, // 13 - Angle, Cycle time

  {   93,  15,  15,   4,   1, }, // 14 - R, P, Y
  {   93,  14,  14,   4,   1, }, // 15 - Thr, Mode

  {   93,  16,  16,   5,   3, }, // 16 - Tri
  {   93,  16,  16,   5,   3, }, // 17 - 
  
  {   93,  18,  18,   5,   3, }, // 18 - Quad+
  {   93,  18,  18,   5,   3, }, // 19

  {   93,  20,  20,   5,   3, }, // 20 - QuadX
  {   93,  20,  20,   5,   3, }, // 21

  {   93,  22,  22,   5,   3, }, // 22 - Bi
  {   93,  22,  22,   5,   3, }, // 23

  {   93,  24,  24,   5,   3, }, // 24 - Gimbal
  {   93,  24,  24,   5,   3, }, // 25

  {   93,  26,  26,   5,   3, }, // 26 - Y6
  {   93,  26,  26,   5,   3, }, // 27

  {   93,  28,  28,   5,   3, }, // 28 - Hex
  {   93,  28,  28,   5,   3, }, // 29

  {   93,  13,  11,   3,   1, }, // 30 - Altitde, Heading

  {   85, UP0,   1,   0,   0, }, // 31 - Acc, Bar, Mag

  {   34,   5,  33,   0,   0, }, // 32 - Profile
  {   86,  32, DN0,   0,   0, }, // 33 - Version

  {   37,  36,  35,  32,   0, }, // 34 - Load Profile
  {   38,  34,  36,  32,   0, }, // 35 - Save Profile
  {   39,  35,  34,  32,   0, }, // 36 - Clear Profile

  {   90,  87,  88,  34,   0, }, // 37 - Load Profile
  {   91,  87,  88,  35,   0, }, // 38 - Save Profile
  {   92,  87,  88,  36,   0, }, // 39 - Clear Profile

  {   93,  40,  40,   5,   3, }, // 40 - Flying Wing
  {   93,  40,  40,   5,   3, }, // 41

#ifdef LOGGING
  {   43,  33,   0,   0,   0, }, // 42 - Logging
  {   94,  45,  44,  42,   0, }, // 43 - Start Logging   #94
  {  101,  43,  45,  42,   0, }, // 44 - Play LogFile    #101
  {   95,  44,  46,  42,   0, }, // 45 - Delete LogFile  #95
  {   96,  45,  43,  42,   0, }, // 46 - Delay time      #96
#endif
  // Right Up  Down Left  Status      Status Flag: 1=Refresh Data, 2=2-Line Mode
};

uint8_t defaultProfile[] = { 'C', 40, 30, 13, 40, 30, 13, 80, 0, 0, 45, 65, 100, 0, 0, 0, };

// configuration-dependent substates of state 'ESC out'
PROGMEM prog_char nextOutState[NUM_CONFIGS] = { 5, 16, 18, 20, 22, 24, 26, 28, 40, };

// charset used in profile names
static char      charsInName[] = " .01234567890_ABCDEFGHIJKLMNOPQRSTUVWXYZ";
static char      profileMsg [] = " #00: 01234567  ";
static char      lineBuffer [17];

// ----------------

static uint8_t   ptr;
static uint8_t   buffer[80];                 // max. number of bytes read from serial

// ----------------

static int8_t    state                ;      // current ui state
static int8_t    profile              ;      // current profile index
static uint8_t   vers                 = 0;   // MultiWii Software Version 1.5
static uint8_t   backlight            = 100; // backlight intensity (0..255)
static boolean   offline              ;      // offline mode when first read is canceled
static uint8_t   minMaxMode           ;      // show act, min or max values
static boolean   graphMode            ;      // graphical output of servo signals
static boolean   charsInitialized     ;      // chars for servo signals are initialized
static int8_t    logMode              ;      // logging mode: -1 not available, 0 - disabled, 1 - enabled
#ifdef LOGGING
static uint16_t  lineNo               ;      // current record number in logFile
static uint8_t   logDelay             = 10;  // x10 in milliseconds
#endif
static uint16_t  keyPressedTime;

// MWC data
static uint8_t   nunchukPresent       ;
static uint8_t   i2C_AccPresent       ;
static uint8_t   i2C_BarometerPresent ;
static uint8_t   i2C_MagnetoPresent   ;
static uint8_t   levelModeParam       ; // if level mode is a activated on the radio : channel 5 value superior to 1700
static uint16_t  cycleTime            ; // this is the number in micro second to achieve a full loop, it can differ a little and is taken into account in the PID loop
static int8_t    config               ; // Tri, Quad, Hex, etc.
static uint8_t   P8[3], I8[3], D8[3]  ; // 8 bits is much faster and the code is much shorter
static uint8_t   rcRate8              ; // dual rate
static uint8_t   rcExpo8              ; // expo
static uint8_t   accStrength8         ; // AutoLevel Strength 0-250
static uint8_t   rollPitchRate        ; // reduces "P" and "D" of Roll/Pitch depending on the associated stick deviation
static uint8_t   yawRate              ; // reduces "P" and "D" of Yaw depending on Rudder stick deviation
static uint8_t   dynThrPID            ; // reduces "P" and "D" of Roll/Pitch when Throttle stick is above 50%

#ifndef MINMAX
static int16_t   accSmooth   [3]      ;  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3]      ;  // from WMP+
static int16_t   altitude             ;  // from barometer
static int16_t   heading              ;  // from compas
static int16_t   angle       [2]      ;  // absolute angle inclination in deg (when Nunchuk/Acc present)
static int16_t   servo       [4]      = { 1500, 1500, 1500, 1500, };             // depends on configuration
static int16_t   motor       [6]      = { 1500, 1500, 1500, 1500, 1500, 1500, }; // depends on configuration 
static int16_t   rcHysteresis[5]      = { 1500, 1500, 1500, 1500, 1500, };       // receiver input
#else
#define TRIPLE  {   0, 32767, -32767}
#define TRIPLEC {1500, 32767, -32767}

static int16_t   accSmooth   [3][3]   = { TRIPLE, TRIPLE, TRIPLE, };  // projection of smoothed and normalized gravitation force vector on x/y/z axis, as measured by accelerometer       
static int16_t   gyroData    [3][3]   = { TRIPLE, TRIPLE, TRIPLE, };  // from WMP+
static int16_t   altitude       [3]   =   TRIPLE;                     // from barometer
static int16_t   heading        [3]   =   TRIPLE;                     // from compas
static int16_t   angle       [2][3]   = { TRIPLE, TRIPLE, };          // absolute angle inclination in deg (when Nunchuk/Acc present)
static int16_t   servo       [4][3]   = { TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, };                   // depends on configuration
static int16_t   motor       [6][3]   = { TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, }; // depends on configuration 
static int16_t   rcHysteresis[5][3]   = { TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, TRIPLEC, };          // receiver input
#endif

// ----------------

#ifdef BACKLIGHT 
  void setBacklight()  {analogWrite(BACKLIGHT_PIN, backlight);}
#else
  #define setBacklight()
#endif

// ----------------

#ifdef STATUS_LEDS
  void green (boolean a)  {digitalWrite( GREEN_LED, a);}
  void yellow(boolean a)  {digitalWrite(YELLOW_LED, a);}
  void red   (boolean a)  {digitalWrite(   RED_LED, a);}
#else
  #define green(a)
  #define yellow(a)
  #define red(a)
#endif

// ----------------

#ifdef JOYSTICK
static int16_t jsHorCtr, jsVertCtr;
#endif

// ----------------

void setup() {
  // make sure that the default chip select pin is set to
  // output, even if you don't use it:
  pinMode(10, OUTPUT);
  
#ifndef SRLCD
  lcd.begin(16, 2);   //no begin in ShiftRegLCD...
#endif
  lcd.clear();

  backlight = 255 - EEPROM.read(5);
  setBacklight();
    
#ifdef TEST
  lcdPrint(12, 1, "Test");
  randomSeed(4711);
#endif

#ifdef JOYSTICK
  pinMode(JS_SELECT_PIN, INPUT);
  pinMode(JS_HOR_PIN,    INPUT);
  pinMode(JS_VERT_PIN,   INPUT);
  
  digitalWrite(JS_SELECT_PIN, HIGH);
  digitalWrite(JS_HOR_PIN,    HIGH);
  digitalWrite(JS_VERT_PIN,   HIGH);
  
  jsHorCtr  = analogRead(JS_HOR_PIN);
  jsVertCtr = analogRead(JS_VERT_PIN);
#endif

#ifdef STATUS_LEDS
  pinMode(   RED_LED, OUTPUT);
  pinMode(YELLOW_LED, OUTPUT);
  pinMode( GREEN_LED, OUTPUT);
#endif

  for (int8_t i = 0; i < 32; i++) {
    lcdPrintProg(max(16-i,0), 0, intro + max(i-16, 0));
    delay((i == 15)? 750 : 50);
  }
  
  if (!restoreParams()) {
    enterVersion();
  }
  setBacklight();

  Serial.begin(SERIAL_COM_SPEED);

#ifdef LOGGING
  // see if the card is present and can be initialized:
  if (!SD.begin(CHIP_SELECT)) {
    setBacklight();
    lcdPrintProg1(noCard);
    logMode = LOG_ERROR;
    storeParams();
    blinkLed(RED_LED, 500, 3);
  }
  setBacklight();
#endif

  state = 31; // eliminates the need to wait for the MWC board to initialize

  if (logMode <= NO_LOGGING) {
    lcdPrintProg(msgs[state], reading);
  
    int readBytes = serialRead(true);
    lcdPrintProg1(readMsg);
    format(6, 1, readBytes, 2, 0);

    blinkLed(YELLOW_LED, 500);
    showStatusLine();
  }
#ifdef LOGGING
  else {
    openLog();
  }
#endif
}

void loop() {
//#ifdef TEST
//  for (int8_t i = 0; i < 5; i++) {rcHysteresis[i] MM_ACT = (rcHysteresis[i] MM_ACT - 1000 + 20) % 1000 + 1000;} // receiver input
//  for (int8_t i = 0; i < 2; i++) {angle       [i] MM_ACT = (angle       [i] MM_ACT + 1 + 2 * i) % 360;}
//#endif      
  
#ifdef LOGGING
  if (logMode > NO_LOGGING) {
    doLog();
    return;
  }
#endif  

  uint8_t statusFlags = getNextState(state, STATUS);
  if (statusFlags & 1) {
    // update the current values from MultiWii board and refresh display
    if (state != 0) {
      serialRead(false);
      showStatusLine();
    }
    else if (nunchukPresent || i2C_AccPresent) {
      serialRead(false);
      // refresh Autolevel state
      lcdPrintProg(8, 1, modeMsg[levelModeParam]);
    }
  }

  int8_t key = getKeyPressed();
  if (key >=0) {
    int8_t next = getNextState(state, key);
#ifdef LOGGING
    if (next == 42 && logMode == LOG_ERROR) {
      next = (key == KEY_UP)? 33 : 0; // disable state 42 "Logging" when initialisation of SD Card failed 
    }
#endif
    
#ifdef TEST
    if (key == KEY_SELECT && state == 5) {
      config = (config + 1) % NUM_CONFIGS;
    }
#endif      

    if (next >= 80) { // handle pseudo-states
      switch (next) {
        case 80:                                 // PID roll
        case 81:                                 // PID pitch
        case 82: { enterPid  (next-80); break; } // PID yaw

        case 83:                                 // dual rate, expo, autolevel strength
        case 84: { enterRates(next-83); break; } // roll/pitch rate, yaw rate, throttle pid

        case 85: { resetProfile();                   state =  0; break; } // reset to default
        case 86: { enterVersion(); enterBacklight(); state = 33; break; } // select Version of MWC software and backlight intensity

        case 87: { profile = (profile+15) % 16;     break; } // decrement profile
        case 88: { profile = (profile+ 1) % 16;     break; } // increment profile

        case 90: { state = restoreProfile()? 34:37; break; } // load profile
        case 91: { storeProfile(); state = 35;      break; } // save profile
        case 92: { deleteProfile();                 break; } // clear profile

        case 93: {
#ifdef GRAPHIC
          if (key == KEY_SELECT || keyPressedTime > 5) {
            graphMode = !graphMode;
//            lcd.clear();
//            lcdPrintProg( 1, 0, graphicMsg);
//            lcdPrintProg(10, 0, onOffMsg[graphMode]);
//            blinkLed(YELLOW_LED, 750);
          }
          else
#endif
          {
#ifdef MINMAX
            minMaxMode = (minMaxMode + 1) % 3; 
            showMinMax();
#endif
          }
          break; 
        }

#ifdef LOGGING
        case 94: {                                      // start logging
          if (!offline) {
  #ifdef LOGVIEW
            logMode = confirm(formatMsg, logFormats)? LF_LOGVIEW : LF_BINARY;
  #else
            logMode = LF_BINARY;
#endif
            boolean flag = confirm(haltReqMsg);
            if (flag) {
              storeParams();
              setError(logEnbMsg);
            }
            openLog();
          }
          else {
            blinkLed(RED_LED, 250, 2);
          }
          break; 
        }
        case 95: { deleteLogFile(); break; }             // delete logfile
        case 96: { enterLogDelay(); break; }             // delay time
        case 101: { playbackLog(); state = 44; break; }  // play logfile
#endif

        case 99: { state = pgm_read_byte_near(nextOutState + getConfig()); break; } // select state depending on config
      }
      next = state;
    }
    
    if (next >= 0 && logMode <= NO_LOGGING) {
      state = next;

      prog_char* line2 = (getNextState(state, STATUS) & 2)? msgs[state+1] : emptyLine;
      lcdPrintProg(msgs[state], line2);
      showStatusLine();
    }
  }
}

int8_t getNextState(int8_t state, int8_t key) { 
  int8_t col = (key == KEY_SELECT)? KEY_RIGHT : (key == STATUS)? STATUS-1 : key;
#ifdef MINMAX
  if (state >= 3 && state <= 5 && col == KEY_RIGHT) {
    showMinMax();
  }
#endif  
  return pgm_read_byte_near(nextState[state] + col); 
}

int8_t getConfig() { 
  return (config < 0 || config >= NUM_CONFIGS)? 0 : config; 
}

// ----------------

void setError(prog_char const msg[]) {
  lcdPrintProg0(msg);
  uint8_t offs = 0;
  while (true) {
    lcdPrintProg1(haltMsg[offs]);
    offs = 1 - offs;
    blinkLed(RED_LED, 2000);
  }
}

void blinkLed(uint8_t led, uint16_t del) {
#ifdef STATUS_LEDS
  del /= 2;
  digitalWrite(led, HIGH); 
  delay(del); 
  digitalWrite(led, LOW); 
#endif
  delay(del);    
}

void blinkLed(uint8_t led, uint16_t del, int8_t repeat) {
  while (repeat-- > 0) {
    blinkLed(led, del);
  }
}

void delay50 () {delay( 50);}

// ----------------

void lcdPrint(int8_t col, int8_t row, char const msg[]) {
  lcd.setCursor(col, row);
  lcd.print(msg);
}

void lcdPrintProg(prog_char const line1[], prog_char const line2[]) {
  lcdPrintProg0(line1);
  lcdPrintProg1(line2);
}

void lcdPrintProg0(prog_char const line[]) {
  lcdPrintProg(0, 0, line);
}

void lcdPrintProg1(prog_char const line[]) {
  lcdPrintProg(0, 1, line);
}

void lcdPrintProg(int8_t col, int8_t row, prog_char const msg[]) {
  char* lptr = lineBuffer;
  char last  = -1;
  for (int8_t i = 0; last != '\0' && i < 16; i++) {
    last    = pgm_read_byte_near(msg + i);
    *lptr++ = last;
  }
  *lptr++ = '\0';

  lcd.setCursor(col, row);
  lcd.print(lineBuffer);
}

void lcdWrite(int8_t col, int8_t digit) {
  lcd.setCursor(col, 1);
  lcd.write(digit + '0');
}

// ----------------

void showStatusLine() {
  switch (state) {
    case 0: {
      lcdPrintProg1(configs[getConfig()]);
      lcd.write(' ');
      lcdPrintProg( 8, 1, modeMsg[offline? 2 : levelModeParam]);
      break;
    }
    case 31: {
      if (keyPressedTime > 10) {
        lcdPrintProg1(metaData);
        format(14, 1, vers, 2, 0);    
      }
      else {
        lcdPrintProg1(statusMsg);
        if (i2C_AccPresent) {
          lcdPrintProg1(accMsg);
        }
        lcdWrite( 3, i2C_AccPresent || nunchukPresent);
        lcdWrite( 9, i2C_BarometerPresent            );
        lcdWrite(15, i2C_MagnetoPresent              );
      }
      break;
    }
    
    case 6:   // PID Roll
    case 7:   // PID Pitch
    case 8: { // PID Yaw
      format      ( 2, 1,  P8[state-6], 4, 2);    
      format      ( 7, 1,  I8[state-6], 5, 4);    
      formatSigned(13, 1, -D8[state-6], 3, 0);    
      break;
    }
    
    case 9: { // Dual Rate, Expo, AutoLevel Strength
      format( 0, 1, rcRate8 * 2 );
      format( 6, 1, rcExpo8     );
      format(12, 1, accStrength8);
      break;
    }
    case 10: { // R/P Rate, Yaw Rate, Throttle PID
      format( 0, 1, rollPitchRate);
      format( 6, 1, yawRate      );
      format(12, 1, dynThrPID    );
      break;
    }
    
    case 11: { // Gyro
      for (int8_t i = 0; i < 3; i++) { formatSigned(i*6, 1, gyroData[i] MM_VAL); }
      break;
    }
    case 12: { // Acc
      for (int8_t i = 0; i < 3; i++) { formatSigned(i*6, 1, accSmooth[i] MM_VAL); }
      break;
    }
    case 13: { // Angles, Cycle
      if (!graphMode) {
        formatSigned( 0, 1, angle[0] MM_VAL);
        formatSigned( 6, 1, angle[1] MM_VAL);
      }
#ifdef GRAPHIC
      else {
        drawAngle(0);
        drawAngle(1);
      }
#endif
      format(12, 1, cycleTime);
      break;
    }

    case 14: { // Roll, Pitch, Yaw
      formatScaledSym( 0, 1, rcHysteresis[ROLL    ] MM_VAL); 
      formatScaledSym( 6, 1, rcHysteresis[PITCH   ] MM_VAL);
      formatScaledSym(12, 1, rcHysteresis[YAW     ] MM_VAL);
      break;
    }
    case 15: { // Throttle, Mode
      formatScaledAbs( 6, 1, rcHysteresis[THROTTLE] MM_VAL);
      formatScaledSym(12, 1, rcHysteresis[AUX1    ] MM_VAL);
      break;
    }

    case 16: { // Left, Right, Rudder, Rear
      formatScaledAbs( 3, 0, motor[2] MM_VAL);
      formatScaledAbs(12, 0, motor[1] MM_VAL);

      formatScaledSym( 3, 1, servo[0] MM_VAL);
      formatScaledAbs(12, 1, motor[0] MM_VAL);
      break;
    }

    case 18: { // Front, Rear, Left, Right
      formatScaledAbs( 3, 0, motor[3] MM_VAL);
      formatScaledAbs(12, 0, motor[0] MM_VAL);

      formatScaledAbs( 3, 1, motor[2] MM_VAL);
      formatScaledAbs(12, 1, motor[1] MM_VAL);
      break;
    }

    case 20: { // Front Left, Front Right, Rear Left, Rear Right
      formatScaledAbs( 3, 0, motor[3] MM_VAL);
      formatScaledAbs(12, 0, motor[1] MM_VAL);

      formatScaledAbs( 3, 1, motor[2] MM_VAL);
      formatScaledAbs(12, 1, motor[0] MM_VAL);
      break;
    }

    case 22: { // Motor 0, Motor 1, Servo 0, Servo 1 
      formatScaledAbs( 3, 0, motor[0] MM_VAL);
      formatScaledAbs(12, 0, motor[1] MM_VAL);

      formatScaledSym( 3, 1, servo[0] MM_VAL);
      formatScaledSym(12, 1, servo[1] MM_VAL);
      break;
    }

    case 24: { // Roll, Pitch
      formatScaledSym( 7, 0, servo[1] MM_VAL);
      formatScaledSym( 7, 1, servo[0] MM_VAL);
      break;
    }

    case 26: { // Up (Left, Mid/Rear, Right), Down (Left, Mid, Right)
      int8_t v5 = (vers <= V1_6_prebis)? 5 : 4; // covers a bug

      formatScaledAbs( 2, 0, motor[2] MM_VAL);
      formatScaledAbs( 7, 0, motor[0] MM_VAL);
      formatScaledAbs(12, 0, motor[1] MM_VAL);

      formatScaledAbs( 7, 1, motor[3   ] MM_VAL);
      formatScaledAbs( 2, 1, motor[9-v5] MM_VAL);
      formatScaledAbs(12, 1, motor[v5  ] MM_VAL);
      break;
    }

    case 28: { // Front (Left, Mid, Right), Rear (Left, Mid, Right)
      int8_t v5 = (vers <= V1_6_prebis)? 5 : 4; // covers a bug

      formatScaledAbs( 2, 0, motor[3 ] MM_VAL);
      formatScaledAbs( 7, 0, motor[v5] MM_VAL);
      formatScaledAbs(12, 0, motor[1 ] MM_VAL);

      formatScaledAbs( 2, 1, motor[2   ] MM_VAL);
      formatScaledAbs( 7, 1, motor[9-v5] MM_VAL);
      formatScaledAbs(12, 1, motor[0   ] MM_VAL);
      break;
    }

    case 30: { // Alt, Heading
      formatSigned( 0, 1, altitude MM_VAL);
      formatSigned( 6, 1, heading  MM_VAL);
      break;
    }
    
    case 37:
    case 38:
    case 39: {
      profileMsg[2] = ('0' + profile/10);
      profileMsg[3] = ('0' + profile%10);
      getProfileName(profileMsg+6);
      
      lcdPrint(0, 1, profileMsg);
      break;
    }

    case 40: { // Servo 0-3 
      formatScaledSym( 3, 0, servo[0] MM_VAL);
      formatScaledSym(12, 0, servo[1] MM_VAL);

      formatScaledSym( 3, 1, servo[2] MM_VAL);
      formatScaledSym(12, 1, servo[3] MM_VAL);
      break;
    }
  }
}

// ----------------

void enterPid(int8_t index) {
  int val[3], minv[3] = { 0, 0, -40 }, maxv[3] = { 200, 250, 0 };

  val[0] =  P8[index]; 
  val[1] =  I8[index]; 
  val[2] = -D8[index];
  
  int8_t local   = 0; // local state (-1..4)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;

  yellow(true);
  while (local >= 0 && local < 3) {
    boolean update = false;

    int8_t key = getKey();
    if (key < 0) {
      lastKey = key;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_LEFT  : { 
          local--;
          update = true; 
          break; 
        }
        case KEY_RIGHT : { 
          local++; 
          update = true; 
          if (local < 3) { // handle RIGHT in third column as SELECT
            break;
          }
        }
        case KEY_SELECT: { 
          lcdPrintProg1(writing);
          
          P8[index] =  val[0];
          I8[index] =  val[1];
          D8[index] = -val[2];
          
          serialWrite();
          local = -1;
          break; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_UP) {
              val[local] = max(val[local]-1, minv[local]);
            }
            else {
              val[local] = min(val[local]+1, maxv[local]);
            }
            switch (local) {
              case 0: { format      ( 2, 1, val[0], 4, 2); break; }
              case 1: { format      ( 7, 1, val[1], 5, 4); break; }    
              case 2: { formatSigned(13, 1, val[2], 3, 0); break; } 
            }
            repeat++;
          }
          break; 
        }
      }
    }
    
    if (update || (count % 20) == 0) {
      char c[3] = {' ', ' ', ' '};
      if ((update || (count / 20) % 2 == 0) && (local >= 0 && local < 3)) {
        c[local] = '\176';
      }
      for (int i = 0; i < 3; i++) {
        lcd.setCursor(max(1, i * 6), 1);
        lcd.write(c[i]);
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
  yellow(false);
}

void enterRates(int8_t index) {
  static char leftCursor2 [] = "\177 ";
  static char rightCursor2[] = " \176";
  static char blank2      [] = "  ";
  
  uint8_t val[3];
 
  if (index == 0) {
    val[0] = rcRate8;
    val[1] = rcExpo8;
    val[2] = accStrength8;
  }
  else {
    val[0] = rollPitchRate;
    val[1] = yawRate;
    val[2] = dynThrPID;
  }
  
  int8_t local   = 0;  // local state (-1..4)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;

  yellow(true);
  while (local >= 0 && local < 3) {
    boolean update = false;
    
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_LEFT  : { 
          local--;
          update = true;
          break; 
        }
        case KEY_RIGHT : { 
          local++; 
          update = true;
          if (local < 3) { // handle RIGHT in third column as SELECT
            break;
          }
        }
        case KEY_SELECT: { 
          lcdPrintProg1(writing);
          
          if (index == 0) {
            rcRate8       = val[0];
            rcExpo8       = val[1];
            accStrength8  = val[2];
          }
          else {
            rollPitchRate = val[0];
            yawRate       = val[1];
            dynThrPID     = val[2];
          }
          
          serialWrite();
          local = -1;
          break; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_UP) {
              uint8_t minVal = (index == 0 && local == 0)? 10 : 0;
              val[local] = max(val[local]-1, minVal);
            }
            else {
              uint8_t maxVal = (index == 0 && local != 1)? 250 : 100;
              val[local] = min(val[local]+1, maxVal);
            }
            uint16_t out = val[local];
            if (index == 0 && local == 0) {
              out *= 2;
            }
            format(local * 6, 1, out);
            repeat++;
          }
          break; 
        }
      }
    }
    
    if (update || (count % 20) == 0) {
      char* left  = blank2;
      char* right = blank2;
      
      if (update || (count / 20) % 2 == 0) {    
        switch (local) {
          case 0: { left = leftCursor2;                        break; }
          case 1: { left = rightCursor2, right = leftCursor2;  break; }
          case 2: {                      right = rightCursor2; break; } 
        }
      }
      lcdPrint( 4, 1, left);
      lcdPrint(10, 1, right);
    }
    count++;
    delay50();
  }
  waitForKeyUp();
  yellow(false);
}

boolean enterName(char name[]) {
  int8_t local   = 0; // local state (-1..8)
  int8_t lastKey = -1;
  int    count   = 0;
  int    repeat  = 0;
  
  for (int8_t i = 0; i < 8; i++) {
    if (name[i] == '-') {
      name[i] = ' ';
    }
  }
  
  int8_t idx = findChar(name[local]);
  
  lcd.cursor();
  lcd.setCursor(local + 6, 1);
  lcd.print(name[local]);
  lcd.setCursor(local + 6, 1);
  yellow(true);
  
  while (local >= 0 && local < 8) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      boolean refresh = false;
      switch (key) {
        case KEY_LEFT: { 
          local--;
          refresh = (local >= 0);
          if (refresh) {
            idx = findChar(name[local]);
          }
          break; 
        }
        case KEY_RIGHT: { 
          local++; 
          refresh = (local < 8);
          if (refresh) { // handle RIGHT in third column as SELECT
            idx = findChar(name[local]);
          }
          break;
        }
        case KEY_SELECT: {
          local = 8;
          break; 
        }
        case KEY_DOWN: case KEY_UP: { 
          if (repeat > 3 || (count % 8) == 0) {
            if (key == KEY_DOWN && idx < sizeof(charsInName)-2) {
              idx++;
              refresh = true;
            }
            else if (key == KEY_UP && idx > 0) {
              idx--;
              refresh = true;
            }
            if (refresh) {
              name[local] = charsInName[idx];
            }
          }
          break; 
        }
      }
      if (refresh) {
        lcd.setCursor(local + 6, 1);
        lcd.print(name[local]);
        lcd.setCursor(local + 6, 1);
        repeat++;
      }
    }
    count++;
    delay50();
  }
  waitForKeyUp();
  yellow(false);
  lcd.noCursor();
  
  return (local == 8);
}

int findChar(char ch) {
  int res = -1;
  for (int8_t i = 0; res < 0 && i < sizeof(charsInName)-1; i++) {
    if (charsInName[i] == ch) {
      res = i;
    }
  }
  if (res < 0) {
    res = 0;
  }
  return res;
}

void enterVersion() {
  yellow(true);
  lcdPrintProg0(versMsg);

  uint8_t val = max(vers, 0);
  int8_t  key = -1;
  while (key < 0 || key == KEY_UP || key == KEY_DOWN) {
    if (key == KEY_DOWN) {
      val = min(val+1, sizeof(frameLength)-1);
    }
    else if (key == KEY_UP) {
      val = max(val-1, 0);
    }
    lcdPrintProg(5, 1, versions[val]);
    delay50();
    key = getKeyPressed();
  }
  if (key != KEY_LEFT) {
    vers = val;
    storeParams();
  }
  yellow(false);
}

void enterBacklight() {
#ifdef BACKLIGHT
  yellow(true);
  lcd.clear();
  lcdPrintProg0(bLightMsg);

  uint8_t val = backlight;
  format(6, 1, val, 3, 0);

  int    repeat  = 0;
  int    count   = 0;
  int8_t lastKey = -1;

  while (true) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_RIGHT: case KEY_SELECT: {
          backlight = val;
          storeParams();
          // intended fall through to KEY_LEFT
        }
        case KEY_LEFT: { 
          waitForKeyUp();
          setBacklight();
          yellow(false);
          return; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            val = (key == KEY_UP)? max(val-5, 0) : min(val+5, 250);
            analogWrite(10, val);
            format(6, 1, val, 3, 0);
            repeat++;
          }
          break; 
        }
      }
    }
    count++;
    delay50();
  }
#endif
}

#ifdef LOGGING
void enterLogDelay() {
  lcdPrint(10, 0, ":");
  format  ( 5, 1, logDelay, 3, 0);
  lcdPrint( 8, 1, "0 ms");
  
  int     val     = logDelay;
  int     repeat  = 0;
  int     count   = 0;
  int8_t  lastKey = -1;

  yellow(true);
  while (true) {
    int8_t key = getKey();
    if (key < 0) {
      lastKey = -1;
      repeat  = 0;
    }
    else if (key == KEY_UP || key == KEY_DOWN || key != lastKey) { // no repeat for LEFT, RIGHT, SEL
      if (key != lastKey) {
        repeat = 0;
      }
      lastKey = key;
      switch (key) {
        case KEY_RIGHT: case KEY_SELECT: {
          logDelay = val;
          storeParams();
          // intended fall through to KEY_LEFT
        }
        case KEY_LEFT: { 
          waitForKeyUp();
          yellow(false);
          return; 
        }
        case KEY_UP: case KEY_DOWN: { 
          if (repeat > 3 || (count % 8) == 0) {
            val = (key == KEY_UP)? max(val-5, 0) : min(val+5, 250);
            format(5, 1, val, 3, 0);
            repeat++;
          }
          break; 
        }
      }
    }
    count++;
    delay50();
  }
}
#endif

// ----------------

void format(int8_t col, int8_t row, int val) {
  format(col, row, val, 4, 0);
}

void format(int8_t col, int8_t row, int val, int8_t len, int8_t prec) {
  formatNum(col, row, val, len, prec, false);
}

void formatSigned(int8_t col, int8_t row, int val) {
  formatSigned(col, row, val, 4, 0);
}

void formatSigned(int8_t col, int8_t row, int val, int8_t len, int8_t prec) {
  formatNum(col, row, val, len, prec, true);
}

void formatNum(int8_t col, int8_t row, int val, int8_t len, int8_t prec, bool sgn) {
  char temp[len+1];
  int8_t low = 0;
  if (sgn) {
    temp[0] = (val > 0)? '+' : (val < 0)? '-' : ' ';
    low++;
  }
  val = abs(val);
  for (int8_t i = len-1; i >= low; i--) {
    if (len-i == prec) {
      temp[i] = '.';
    }
    else {
      temp[i] = '0' + (val % 10);
      val    /= 10;
    }
  }
  if (val > 0) { // overflow
    for (int8_t i = len-1; i >= 0; i--) {
      temp[i] = '#';
    }
  }
  temp[len] = '\0';
  lcdPrint(col, row, temp);
}

void formatScaledSym(int8_t col, int8_t row, int val) { // -100..+100
  val = (val - 1500)/5;
  if (!graphMode) {
    formatSigned(col, row, val);
  }
#ifdef GRAPHIC
  else {
    formatGraphic(col, row, (val+100)/2);
  }
#endif
}

void formatScaledAbs(int8_t col, int8_t row, int val) { // 0..100
  val = (val - 1000)/10;
  if (!graphMode) {
    formatSigned(col, row, val);
  }
#ifdef GRAPHIC
  else {
    formatGraphic(col, row, val);
  }
#endif
}

#ifdef GRAPHIC
void formatGraphic(int8_t col, int8_t row, int val) { // 0..100
  if (!charsInitialized) {
    charsInitialized = true;
  
    static byte bars[8] = { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x15, };
    static byte less[8] = { 0x00, 0x04, 0x0C, 0x1C, 0x0C, 0x04, 0x00, 0x15, };
    static byte grt [8] = { 0x00, 0x04, 0x06, 0x07, 0x06, 0x04, 0x00, 0x15, };

    byte pattern = 0x10;
    for (int8_t i = 0; i <= 5; i++) {
      for (int8_t j = 0; j < 7; j++) {
        bars[j] = pattern;
      }
      lcd.createChar(i, bars);
      pattern >>= 1;
    }
    lcd.createChar(6, less);
    lcd.createChar(7, grt);
  }  
  
  char bar[] = { 5, 5, 5, 5 };
  
  if      (val < -100 || val > 200) { bar[0] = 6; bar[3] = 7; } // invalid
  else if (val <    0) { bar[0]      = 6; }           // <...
  else if (val >= 100) { bar[3]      = 7; }           // ...>
  else                 { bar[val/25] = (val%25)/5; }  // ..|.
  lcd.setCursor(col, row);
  lcd.write(bar[0]); 
  lcd.write(bar[1]); 
  lcd.write(bar[2]); 
  lcd.write(bar[3]); 
}

/* ----- */

static uint8_t angleChars[4][8];

void drawAngle(int8_t index) {

  // clear char arrays
  uint8_t *cptr = angleChars[0];
  for (int8_t i = 0; i < 32; i++) {
    *cptr++ = 0;
  }

  int8_t charOffset = index * 4;
  int16_t alpha = -angle[index] MM_VAL;
  if (alpha >= -360 && alpha <= 360) {
    float   ang   = radians(alpha);
    float   sinA  = sin(ang);
    float   cosA  = cos(ang);
    
  //  if (abs(sinA) > 0.25f) {
  //    angleChars[0][4] = 0x14;  // mark horizon
  //    angleChars[3][4] = 0x05;
  //  }
    
    int8_t x0 = (int8_t) (- cosA * 10.0f + 0.5f);
    int8_t y0 = (int8_t) (- sinA * 10.0f + 0.5f);
    
    drawLine(charOffset, x0, y0, -x0, -y0);
    
    int8_t x1 = (int8_t) (- sinA * 3.0f + 0.5f);
    int8_t y1 = (int8_t) (  cosA * 3.0f + 0.5f);
  
    drawLine(charOffset, 0, 0, x1, y1);
  }
  else {
    angleChars[1][4] = 0x06;  
    angleChars[2][4] = 0x18;  
  }
  
  for (int8_t i = 0; i < 4; i++) {
    lcd.createChar(charOffset + i, angleChars[i]);
    lcd.setCursor (index * 6  + i, 1);
    lcd.write     (charOffset + i);
  }
  
  charsInitialized = false;  
}

void drawLine(int8_t charOffset, int8_t x0, int8_t y0, int8_t x1, int8_t y1) {
  // Bresenham again...
  
  int8_t dx =  abs(x1-x0), sx = x0<x1 ? 1 : -1;
  int8_t dy = -abs(y1-y0), sy = y0<y1 ? 1 : -1; 
  int8_t err = dx+dy, e2; /* error value e_xy */
 
  for(;;){  /* loop */
    setPixel(x0, y0);
    if (x0==x1 && y0==y1) break;
    e2 = 2*err;
    if (e2 >= dy) { err += dy; x0 += sx; } /* e_xy+e_x > 0 */
    if (e2 <= dx) { err += dx; y0 += sy; } /* e_xy+e_y < 0 */
  }  
}

void setPixel(int8_t x, int8_t y) { 
  // x: -10..+9, y: -3..+4
  
  if (x >= -10 && x < 10 && y >= -3 && y < 5) {
    int8_t nrow = 4 - y;   // 0..7
    int8_t ncol = x + 10;  // 0..19
    int8_t idx  = ncol / 5;
    int8_t pcol = ncol % 5;
    
    angleChars[idx][nrow] |= (0x10 >> pcol);
  }
}
#endif // GRAPHIC

// ----------------

#ifdef MINMAX
void showMinMax() {
  lcd.clear();
  lcdPrintProg(1, 0, minMaxMsg);
  lcdPrintProg(7, 0, minMaxDsc[minMaxMode]);
  blinkLed(YELLOW_LED, 750);
}
#endif

// ----------------

boolean confirm(prog_char const msg[]) {
  return confirm(msg, yesNoMsg);
}

boolean confirm(prog_char const msg1[], prog_char const msg2[]) {
  yellow(true);
  lcdPrintProg(msg1, msg2);
    
  int8_t key = -1;
  while (key != KEY_LEFT && key != KEY_RIGHT) {
    delay50();
    key = getKeyPressed();
  }      

  yellow(false);
  return (key == KEY_RIGHT);
}

int8_t enterKey() {
  yellow(true);
  int8_t key = -1;
  while ((key = getKeyPressed()) < 0);
  yellow(false);
  return key;
}

int8_t getKeyPressed() {
  long start = millis();
  int8_t key = getKey();
  if (key >= 0) {
    waitForKeyUp();
  }
  keyPressedTime = (uint16_t) (millis() - start)/100;
  return key;
} 

void waitForKeyUp() {
  while (rawGetKey() >= 0) {
//    delay50();
  }
}

int8_t getKey() {
  static int8_t oldKey = -1;

  int8_t key = rawGetKey();
  if (key != oldKey) {
    delay50();		// wait for debounce time
    key = rawGetKey();
    if (key != oldKey) {			
      oldKey = key;
    }
  }
  return key;
}

#ifndef HWKEYS
int adc_key_val[NUM_KEYS] = { 30, 150, 360, 535, 760 }; // thresholds of the cascading resistors
#endif

int8_t rawGetKey() {
  int8_t result = -1;

#ifdef HWKEYS
  if      (digitalRead(HW_RIGHT) ==0) {result = KEY_RIGHT; }
  else if (digitalRead(HW_UP)    ==0) {result = KEY_UP;    }
  else if (digitalRead(HW_DOWN)  ==0) {result = KEY_DOWN;  }
  else if (digitalRead(HW_LEFT)  ==0) {result = KEY_LEFT;  }
  else if (digitalRead(HW_SELECT)==0) {result = KEY_SELECT;}
#else
  uint16_t input  = analogRead(0);   // read the value from the sensor  
  for (uint8_t k = 0; result < 0 && k < NUM_KEYS; k++) {
    if (input < adc_key_val[k]) {
      result = k;
    }
  }
#endif

#ifdef JOYSTICK
  if (result < 0) {
    int sel  = analogRead(JS_SELECT_PIN);
    int hor  = analogRead(JS_HOR_PIN);
    int vert = analogRead(JS_VERT_PIN);

    if (sel < 1000 || hor < 1000 || vert < 1000) { // is joystick present? (int. pullups are on)
#ifdef JS_INV_SELECT_KEY
      boolean select = (sel < 3);    // LOW when pressed 
#else
      boolean select = (sel > 1020); // HIGH when pressed
#endif
      if (select) {
        result = KEY_SELECT;
      }
      else {
        hor  = (hor  - jsHorCtr)  * JS_HOR_SIGN;
        vert = (vert - jsVertCtr) * JS_VERT_SIGN;
    
        if (abs(vert) < JS_DEADBAND) {
          if      (hor < -10 * JS_DEADBAND)  {result = KEY_RIGHT;}
          else if (hor >  10 * JS_DEADBAND)  {result = KEY_LEFT; }
        }
        else if (abs(hor) < JS_DEADBAND) {
          if      (vert < -10 * JS_DEADBAND) {result = KEY_UP;   }
          else if (vert >  10 * JS_DEADBAND) {result = KEY_DOWN; }
        }
      }
    }
  }
#endif

  return result;
}

// ----------------

void serialize8 (int8_t  a) {buffer[ptr++] = a;}
void serialize16(int16_t a) {buffer[ptr++] = a; buffer[ptr++] = a/256;}

void deserialize8 (int8_t&   val) {val = buffer[ptr++];}
void deserialize8 (uint8_t&  val) {val = buffer[ptr++];}
void deserialize16(int16_t&  val) {uint8_t low = buffer[ptr++]; uint8_t hi = buffer[ptr++]; val = (int16_t) (hi * 256 + low);}
void deserialize16(uint16_t& val) {uint8_t low = buffer[ptr++]; uint8_t hi = buffer[ptr++]; val = (int16_t) (hi * 256 + low);}
void deserialize16(int16_t arr[]) {deserialize16(arr[0]); arr[1] = min(arr[0], arr[1]); arr[2] = max(arr[0], arr[2]);}

#ifndef TEST
int serialRead(boolean init) {
  int readBytes = 0;
  if (offline) {
    blinkLed(RED_LED, 20);
    delay(100);
  }
  else {
    boolean frameErr  = false;
    uint8_t len       = frameLength[vers];
    while (true) {
      Serial.flush();
    
      boolean avail;
      int     retry = 0;
      do {
        Serial.write('A');
        delay50();
        avail = Serial.available();
        if (!avail) {
          retry++;
          if (retry > 600) {
            setError(timeoutError);
          }
          else if (retry > 10) {
            if (init && getKey() == KEY_LEFT) {
              // cancel loop
              lcdPrintProg1(cancelRead);
              blinkLed(RED_LED, 250, 6);
              offline = true;
//              logMode = LOG_ERROR;
              return 0;
            }
            blinkLed(RED_LED, 4);
          }
        }
      }
      while (!avail); // 30 seconds until timeout
    
      blinkLed(YELLOW_LED, 4);
        
      char prefix = Serial.read();
      delay(2);
    
      if (prefix == 'A') {
        readBytes = 0;
        buffer[readBytes++] = prefix;
        while (readBytes <= len && Serial.available()) {
          buffer[readBytes++] = Serial.read();
          delay(2);
       }
        if (readBytes > len && buffer[readBytes-1] == 'A') {
          deserialize();
          break;
        }
        else if (!frameErr) {
          frameErr = true;
        }
        else {
          vers = -1;       // set version invalid to force version request after next start
          storeParams();
          setError(frameError);
        }
      }
    }
  }
  return readBytes;
}
#else
void serRnd (int16_t val) {serialize16(val + random(49) - 24);}
void serRndA(int16_t val) {serialize16(val + random(5)  - 2);}

int serialRead(boolean init) {
  ptr = 0;

  serialize8('A');
  for (int8_t i=0; i<3; i++) serRnd(accSmooth[i] MM_ACT);         // acc
  for (int8_t i=0; i<3; i++) serRnd(gyroData [i] MM_ACT);         // gyro

  serialize16(0);                                                 // alt
  if (vers >= V1_6_prebis) {
    serialize16(0);                                               // heading
  }

  int8_t numServos = (vers >= V1_6)? 4 : 2;  
  for (int8_t i=0; i<numServos; i++) serRnd(servo       [i] MM_ACT); // servos
  for (int8_t i=0; i<6;         i++) serRnd(motor       [i] MM_ACT); // motor
  for (int8_t i=0; i<5;         i++) serRnd(rcHysteresis[i] MM_ACT); // receiver

  serialize8(1);                                                  // nunchuk

  if (vers >= V1_6_prebis) {
    serialize8(0);                                                // accPresent
    serialize8(0);                                                // baro
    serialize8(0);                                                // magneto
  }
  
  serialize8(1);                                                  // levelmode
  serialize16(6000);                                              // cycle time
  
  for (int8_t i=0; i < 2; i++) serRndA(angle[i] MM_ACT);          // roll and pitch angle
  
  serialize8(2);                                                  // config
  for (int8_t i=0; i < 3; i++) {serialize8(40); serialize8(30); serialize8(13);} // pid

  serialize8((vers == V1_5)? 100 : 50);                           // dual rate
  serialize8(66);                                                 // expo
  serialize8(111);                                                // accStrength
  serialize8(0);                                                  // roll/pitch rate
  serialize8(0);                                                  // yaw rate
  serialize8(0);                                                  // throttle PID rate
//  serialize8('A');
    
  deserialize();

  blinkLed(YELLOW_LED, 4);
  delay(200);

  return ptr;
}
#endif

void deserialize() {
  ptr = 1;
  
  int8_t i;
  for (i=0; i<3; i++) {deserialize16(accSmooth[i]);}
  for (i=0; i<3; i++) {deserialize16(gyroData [i]);}
  deserialize16(altitude);
  if (vers >= V1_6_prebis) {
    deserialize16(heading);
  }

  int8_t numServos = (vers >= V1_6)? 4 : 2;  
  for (i=0; i<numServos; i++) {deserialize16(servo       [i]);}
  for (i=0; i<6;         i++) {deserialize16(motor       [i]);}
  for (i=0; i<5;         i++) {deserialize16(rcHysteresis[i]);}

  deserialize8(nunchukPresent);

  if (vers >= V1_6_prebis) {
    deserialize8(i2C_AccPresent      );
    deserialize8(i2C_BarometerPresent);
    deserialize8(i2C_MagnetoPresent  );
  }  

  deserialize8(levelModeParam); 
  deserialize16(cycleTime);
  
  deserialize16(angle[0]);
  deserialize16(angle[1]);
 
  deserialize8(config); 

  for (i=0; i<3; i++) {
    deserialize8(P8[i]); 
    deserialize8(I8[i]);
    deserialize8(D8[i]);
  }
  deserialize8(rcRate8      );
  if (vers == V1_5) { rcRate8 /= 2; }
  deserialize8(rcExpo8      ); 
  deserialize8(accStrength8 ); 
  deserialize8(rollPitchRate); 
  deserialize8(yawRate      ); 
  deserialize8(dynThrPID    ); 
}

void serialWrite() {
  ptr = 0;
  serialize8('C');
  for (int8_t i=0; i<3; i++) { serialize8(P8[i]); serialize8(I8[i]); serialize8(D8[i]); }
  uint8_t dualRate = (vers == V1_5)? rcRate8 * 2 : rcRate8;
  serialize8(dualRate     );
  serialize8(rcExpo8      );
  serialize8(accStrength8 );
  serialize8(rollPitchRate);
  serialize8(yawRate      );
  serialize8(dynThrPID    );
  
  Serial.write(buffer, ptr);

  lcdPrintProg1(writeMsg);
  format(7, 1, ptr, 2, 0);

  blinkLed(GREEN_LED, 500, 3); // signal "serial write"
  
  serialRead(false); // reload the frame
}

// ----------------

void resetProfile() {
  if (confirm(resetMsg)) {
    extractProfile(defaultProfile);
//    P8[0] = 40; I8[0] = 30; D8[0] = 13;
//    P8[1] = 40; I8[1] = 30; D8[1] = 13;
//    P8[2] = 80; I8[2] =  0; D8[2] =  0;
//    
//    rcRate8       = 45;
//    rcExpo8       = 65;
//    accStrength8  = 100;
//    rollPitchRate = 0;
//    yawRate       = 0;
//    dynThrPID     = 0;
//    
//    serialWrite();
  }
}

void storeProfile() {
  getProfileName(lineBuffer); 
  if (enterName(lineBuffer)) {
    storeProfile(lineBuffer);
  }
}

boolean restoreProfile() {
  uint16_t adr  = (profile+1) * 32 + 8; // skip name
  uint8_t* lptr = buffer;
  for (int8_t i = 0; i < 16; i++) {
    *lptr++ = EEPROM.read(adr++);
  }
  return extractProfile(buffer);
}
 
boolean extractProfile(uint8_t* lptr) { 
  boolean succ = (*lptr++ == 'C');  
  if (succ) {
    for (int8_t i=0; i<3; i++) {
      P8[i]       = *lptr++; 
      I8[i]       = *lptr++; 
      D8[i]       = *lptr++;
    }
    rcRate8       = *lptr++;
    rcExpo8       = *lptr++;
    accStrength8  = *lptr++;
    rollPitchRate = *lptr++;
    yawRate       = *lptr++;
    dynThrPID     = *lptr++;

    serialWrite();
  }
  return succ;
}

void storeProfile(char* name) {
  uint16_t adr  = (profile+1) * 32;
  uint8_t* lptr = buffer;
  int8_t   i    = 0;

  while (name[i] != '\0' && i < 8) { *lptr++ = name[i++]; }
  while (i < 8)                    { *lptr++ = ' ';       }
  
  *lptr++ = 'C'          ;
  for (i = 0; i < 3; i++) {*lptr++ = P8[i]; *lptr++ = I8[i]; *lptr++ = D8[i];}
  *lptr++ = rcRate8      ;
  *lptr++ = rcExpo8      ;
  *lptr++ = accStrength8 ;
  *lptr++ = rollPitchRate;
  *lptr++ = yawRate      ;
  *lptr++ = dynThrPID    ;
  
  for (i = 0; i < 24; i++) {
    EEPROM.write(adr++, buffer[i]);
  }
}

void deleteProfile() {
  uint16_t adr = (profile+1) * 32;
  for (int8_t i = 0; i < 32; i++) {
    EEPROM.write(adr++, 0);
  }
}

bool getProfileName(char* name) {
  uint16_t adr = (profile+1) * 32;
  for (int8_t i = 0; i < 8; i++) { name[i] = EEPROM.read(adr++); }
  //name[8] = '\0'; 

  char    hdr  = EEPROM.read(adr++);
  boolean succ = (hdr == 'C');  
  if (!succ) {
    for (int8_t i = 0; i < 8; i++) { name[i] = '-'; }
  }
  return succ;
} 

// ----------------

boolean restoreParams() {
  byte    magic0 = EEPROM.read(0); // check validity of data by storing some magic numbers as a prefix
  byte    magic1 = EEPROM.read(1);
  uint8_t tv     = EEPROM.read(2);
  if (tv >= 15 && tv < 16) {
    tv -= 15;
  }
  boolean succ = (magic0 == 47 && magic1 == 11 && tv >= 0 && tv < sizeof(frameLength));

  if (succ) {
    vers    = tv;
#ifdef LOGGING
    logMode   = EEPROM.read(3);
    if (logMode == LOG_ERROR) {logMode = NO_LOGGING;}
    logDelay  = EEPROM.read(4);
#endif
    backlight = 255 - EEPROM.read(5);
  }
  return succ;
}

void storeParams() {
  EEPROM.write(0, 47);  
  EEPROM.write(1, 11);  
  EEPROM.write(2, vers);
#ifdef LOGGING  
  EEPROM.write(3, logMode);
  EEPROM.write(4, logDelay);
#endif
  EEPROM.write(5, 255 - backlight);
}

// ----------------

#ifdef LOGGING
void formatLogFile(int cnt) {
  fileName[5] = (cnt >= 500)? 'v' : 'o';    // .log - binary, .lvw - LogView 
  fileName[6] = (cnt >= 500)? 'w' : 'g';    // (Note: '.lov' Files are _not_ in OpenFormat)
  fileName[2] = '0' + cnt % 10; cnt /= 10;
  fileName[1] = '0' + cnt % 10; cnt /= 10;
  fileName[0] = '0' + cnt % 10;
}

void openLog() {
  boolean created = false; 
  int     cnt     = 0;
#ifdef LOGVIEW
  int     offset  = (logMode == LF_LOGVIEW)? 500 : 0;
#else
  int     offset  = 0;
#endif

  while (!created && cnt < 500) {
    formatLogFile(offset + cnt);
    if (!SD.exists(fileName)) {
      created = SD.open(fileName, FILE_WRITE);
      setBacklight();
    }
    else {
      cnt++;
    }
  }
  if (!created) {
    logMode = NO_LOGGING;
    storeParams();
    setError(fileError);
  }
  state = 0;
  storeParams();

  lcdPrintProg(logMsg, fileMsg);
  format(6, 1, offset + cnt, 3, 0);
  lcd.print(" #");

  green(true);
}

int findNextLogFile(int act, boolean ascend, boolean binOnly) {
  int     mod   = (binOnly)? 500 : 1000;
  int     cnt   = (ascend)? (act + 1) % mod : (act + mod - 1) % mod;
  boolean found = false;
  while (!found && cnt != act) {
    formatLogFile(cnt);
    found = SD.exists(fileName);
    setBacklight();
    if (!found) {
      cnt = (ascend)? (cnt + 1) % mod : (cnt + mod - 1) % mod;
    }
  }
  if (!found) {
    formatLogFile(act);          // check if exactly one logfile exists
    found = SD.exists(fileName);
    setBacklight();
  }
  return (!found)? -1 : cnt;
}

#ifdef LOGVIEW
byte* toBuffer(byte* outPtr, int16_t value) {
  if (value == 0) {
    *outPtr++ = '0';
  }
  else {
    if (value < 0) {
      value = -value;
      *outPtr++ = '-';
    }
    boolean flag = false;
    int16_t dec  = 10000;
    while (dec > 0) {
      int val = (value / dec) % 10;
      flag |= (val > 0);
      if (flag) {
        *outPtr++ = val + '0';
      }
      dec /= 10;
    }
  }
  *outPtr++ = ';';
  return outPtr;
}

byte* toBufferPwm(byte* outPtr, int16_t value, boolean symm) {
  if (value != 0) {
    if (symm) {
      value = (value - 1500) / 5; // symmetric [-100..+100]
    }
    else {
      value = (value - 1000) / 10; // percent (0..100)
    }
    outPtr = toBuffer(outPtr, value);
  }
  else {
    *outPtr++ = ';';
  } 
  return outPtr;
}

byte* addPrefix(byte* outPtr, char channel) {
    *outPtr++ = '$';
    *outPtr++ = channel;
    *outPtr++ = ';';
    *outPtr++ = '1' + levelModeParam;  // state
    *outPtr++ = ';';
    outPtr = toBuffer(outPtr, lineNo); // used as timeStamp
    return outPtr;
}

byte* addSuffix(byte* outPtr) {
    *outPtr++ = '0';  // checksum - not used
    *outPtr++ = '\r';
    *outPtr++ = '\n';
    return outPtr;
}
#endif

//int8_t channels[] = { THROTTLE, ROLL, PITCH, YAW, AUX1, };

void doLog() {
#ifdef LOGVIEW
  if (logMode == LF_LOGVIEW) {
    deserialize();

    byte  outBuffer[160];
    byte* outPtr = outBuffer;

    outPtr = addPrefix(outPtr, '1');
    for (int8_t i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, accSmooth[i] MM_ACT); } // 2 - 4
    for (int8_t i = 0; i < 3; i++) { outPtr = toBuffer(outPtr, gyroData [i] MM_ACT); } // 5 - 7
    outPtr = toBuffer(outPtr, altitude MM_ACT);                                        // 8
    outPtr = toBuffer(outPtr, heading  MM_ACT);                                        // 9
    for (int8_t i = 0; i < 2; i++) { outPtr = toBuffer(outPtr, angle    [i] MM_ACT); } // 10 - 11
    outPtr = addSuffix(outPtr);

    outPtr = addPrefix(outPtr, '2');
    for (int8_t i = 0; i < 4; i++) { outPtr = toBufferPwm(outPtr, servo[i] MM_ACT, true ); } // 1 - 4
    for (int8_t i = 0; i < 6; i++) { outPtr = toBufferPwm(outPtr, motor[i] MM_ACT, false); } // 5 - 10
    outPtr = addSuffix(outPtr);

    outPtr = addPrefix(outPtr, '3');
    for (int8_t i = 0; i < 5; i++) { outPtr = toBufferPwm(outPtr, rcHysteresis[i] MM_ACT, (i>0)); } // 1 - 5
    outPtr = addSuffix(outPtr);
    
    logFile.write(outBuffer, outPtr - outBuffer);
  }
  else
#endif
  {
    logFile.write(buffer, ptr);
  }

  setBacklight();

  lineNo++;
  
  int8_t key = getKeyPressed();
  if (key >= 0) {
    bool flag = confirm(stopLogMsg);
    if (flag) {
      // close the logfile and stop logging
      logFile.close();
      setBacklight();
      green(false);
      lineNo  = 0;
      logMode = NO_LOGGING;
      storeParams();
      waitForKeyUp();

      state = 43;
      lcdPrintProg(msgs[state], emptyLine);
      return;
    }
    else {
      // continue logging
      lcdPrintProg1(fileMsg);
      lcd.setCursor(6, 1);
      lcd.write(fileName[0]);
      lcd.write(fileName[1]);
      lcd.write(fileName[2]);
      lcd.print(" #");
    }
  }
  int readBytes = serialRead(false);
  if (readBytes > 0) {
    format(11, 1, lineNo, 5, 0);
    blinkLed(YELLOW_LED, 10);
    delay(logDelay * 10);
  }
  else {
    logMode = NO_LOGGING;
    storeParams();
  }
}

int16_t selectLogfile(int next, boolean binOnly) {
  boolean cancel = false;
  next = findNextLogFile(next, true, binOnly);
  while (next >= 0) {
    lcdPrintProg(3, 1, fileMsg);
    format(9, 1, next, 3, 0);
    
    int8_t key = enterKey();
    if (key == KEY_UP || key == KEY_DOWN) {
      next = findNextLogFile(next, key == KEY_DOWN, binOnly);
    }
    else if (key == KEY_SELECT || key == KEY_RIGHT) {
      break;
    }
    else if (key == KEY_LEFT) {
      next   = -1;
      cancel = true;
    }
  }
  if (next < 0 && !cancel) {
    blinkLed(RED_LED, 250, 2);
  }
  return next;
}

void playbackLog() {
  int fileNr = selectLogfile(499, true);
  while (fileNr >= 0) {
    green(true);
    lcdPrintProg1(fileMsg);
    lcdPrintProg(6, 1, fileSuff);
    format(6, 1, fileNr, 3, 0);
        
    lineNo = 0;
//    formatLogFile(fileNr);
    logFile = SD.open(fileName, FILE_READ);
    
    uint8_t len = frameLength[vers];
    while (logFile.available()) {
      format(11, 1, lineNo, 5, 0);
      lineNo++;
    
      ptr = 0;
      while (ptr < len && logFile.available()) {
        buffer[ptr++] = logFile.read();
      }
      setBacklight();
      buffer[ptr++] = 'A';
      if (ptr == len+1) {
        int timeout = 600;
        while (timeout > 0 && !Serial.available()) {
          if (timeout <= 590) {
            blinkLed(RED_LED, 4);
          }
          delay50();
          timeout--;
        }
        if (timeout <= 0) {
          lcdPrintProg1(timeoutError);
          blinkLed(RED_LED, 500, 3);
          break;
        }                
        if (Serial.read() == 'A') {
          blinkLed(YELLOW_LED, 4);
          Serial.write(buffer, ptr);
          delay(logDelay * 10); // in real-time
        }
      }
      else {
        break; // eof
      }
    }
    logFile.close();
    green(false);

    lcdPrintProg1(emptyLine);
    fileNr = selectLogfile(fileNr-1, true);
  }
}

void deleteLogFile() {
  int fileNr = selectLogfile(999, false);
  while (fileNr >= 0) {
//    formatLogFile(fileNr);
    boolean succ = SD.remove(fileName);
    setBacklight();

    lcdPrintProg1((succ)? fileDeleted : deleteFailed);
    blinkLed(     (succ)? YELLOW_LED  : RED_LED, 250, 6);
    lcdPrintProg1(emptyLine);

    if (succ) {
      fileNr = selectLogfile(fileNr, false);
    }
  }
}
#endif

